import catboost as cb
import numpy as np
import torch
from torch import nn


class CatModule(nn.Module):
    """
    CatModule mirrors LGBModule and XGBModule but uses catboost as the gbdt
    implementation.

    Currently this is in the experimental section of the code because, while
    this module generally gives results as expected, the catboost connection
    cannot yet work the same as the GBDT connections in LGBModule and
    XGBModule; thus, using CatModule is a fair amount slower compared to the
    other modules. If you have ideas here, please file an issue :)
    """

    def __init__(
        self,
        batch_size,
        input_dim,
        output_dim,
        params={},
    ):
        super(CatModule, self).__init__()
        self.batch_size = batch_size
        self.input_dim = input_dim
        self.output_dim = output_dim

        self.params = params
        self.params.update(
            {
                "iterations": 1,
                "eval_metric": ("MultiRMSE" if self.output_dim > 1 else "RMSE"),
                "boost_from_average": False,
            }
        )

        self.bst = None
        self.train_dat = None
        self.obj = None

        self.FX = nn.Parameter(
            torch.tensor(
                np.zeros([batch_size, output_dim]),
                dtype=torch.float,
            )
        )

    def forward(self, input_array, return_tensor=True):
        assert isinstance(input_array, np.ndarray), "Input must be a numpy array"
        # TODO figure out how actual batch training works here
        if self.training:
            if self.bst and self.train_dat:
                preds = self.bst.predict(self.train_dat)
            else:
                preds = np.zeros([self.batch_size, self.output_dim])
        else:
            if self.bst:
                preds = self.bst.predict(input_array)
            else:
                preds = np.zeros(
                    [input_array.shape[0], self.output_dim], dtype=torch.float
                )

        if self.training:
            FX_detach = self.FX.detach()
            FX_detach.copy_(
                torch.tensor(
                    preds.reshape([self.batch_size, self.output_dim]), dtype=torch.float
                )
            )

        if return_tensor:
            if self.training:
                return self.FX
            else:
                return torch.tensor(
                    preds.reshape([-1, self.output_dim]), dtype=torch.float
                )
        return preds

    def gb_step(self, input_array):
        grad = self.FX.grad * self.batch_size

        # parameters are independent row by row, so we can
        # at least calculate hessians column by column by
        # considering the sum of the gradient columns
        hesses = []
        for i in range(self.output_dim):
            hesses.append(
                torch.autograd.grad(grad[:, i].sum(), self.FX, retain_graph=True)[0][
                    :, i : (i + 1)
                ]
            )
        hess = torch.cat(hesses, axis=1)

        self.grad_index = 0
        if self.bst is not None:
            self.obj.reset(grad, hess, self.batch_size)
            self.bst.fit(self.train_dat, init_model=self.bst)

        else:
            self.obj = CatboostObj(grad, hess, self.batch_size)
            input_params = self.params.copy()
            input_params.update({"loss_function": self.obj, "allow_const_label": True})
            self.train_dat = cb.Pool(
                data=input_array, label=np.zeros([self.batch_size, self.output_dim])
            )
            self.bst = cb.CatBoostRegressor(**input_params)
            self.bst.fit(self.train_dat)


class CatboostObj(cb.MultiTargetCustomObjective):
    def __init__(self, grad, hess, n):
        self.grad = grad.detach().numpy()
        self.hess = hess.detach().numpy()
        self.call_index = 0
        self.n = n

    def reset(self, grad, hess, n):
        self.grad = grad.detach().numpy()
        self.hess = hess.detach().numpy()
        self.call_index = 0
        self.n = n

    def calc_ders_multi(self, approxes, targets, weight):
        # TODO catboost works row by row in the multidim output setting.
        # I'm not sure if we can re-calculate grad and hess per row efficiently,
        # so, currently, this relies on the objective function being called
        # sequentially, a very brittle assumption.
        grad = -self.grad[self.call_index % self.n, :]  # uses the negative gradient
        hess = -np.diag(self.hess[self.call_index % self.n, :])
        self.call_index += 1
        return (grad, hess)



import numpy as np
import pandas as pd
import torch
import torch.nn as nn
from scipy.linalg import lstsq
from scipy.optimize import root_scalar
from sklearn.base import BaseEstimator, RegressorMixin
from sklearn.utils.validation import check_is_fitted

from gbnet.gblinear import GBLinear


def pd_datetime_to_seconds(x: pd.Series):
    return pd.to_numeric(x) / 1000000000.0  # convert to seconds


def loadModule(module):
    assert module in {"XGBModule", "LGBModule"}
    if module == "XGBModule":
        from gbnet import xgbmodule

        return xgbmodule.XGBModule
    if module == "LGBModule":
        from gbnet import lgbmodule

        return lgbmodule.LGBModule


class Forecast(BaseEstimator, RegressorMixin):
    """
    A forecasting model class that implements a trend + seasonality + changepoints
    model using either XGBModule or LGBModule (but defaulting to XGBModule).

    Parameters
    ----------
    nrounds : int, default=50
        Number of training iterations (epochs) for the model.
    params : dict, optional
        Dictionary of additional parameters to be passed to the underlying forecast model.
        Defaults to {"eta": 0.17, "max_depth": 3, "lambda": 1, "alpha": 8}
    module_type : str, default="XGBModule"
        Type of gradient boosting module to use, either "XGBModule" or "LGBModule".
    linear_params : dict, default={}
        Parameters to pass to GBLinear.
        Defaults to {"min_hess": 0.0, "lambd": 0.1, "lr": 0.9}
    changepoint_params : dict, default={}
        Parameters for changepoint detection and modeling. Defaults to:
        {
            "n_changepoints": 32,  # how many changepoints to consider
            "eta": 0.9,
            "max_depth": 9,
            "lambda": 6.5,
            "alpha": 3.8,
            "cp_gap": 0.5,  # portion of time series allowing changepoints
            "cp_train_gap": 4  # how many training rounds to NOT update the periodic component
        }

    Attributes
    ----------
    nrounds : int
        Number of training rounds for the model.
    params : dict
        Additional parameters passed to the forecast model.
    model_ : ForecastModule or None
        Trained forecast model instance. Set after fitting.
    losses_ : list
        List of loss values recorded at each training iteration.

    Methods
    -------
    fit(X, y)
        Trains the forecast model using the input features X and target variable y.
        X must contain the datetime column 'ds'.
    predict(X, components=False)
        Predicts target values based on the input features X.
        If components=True, returns tuple of (trend, periodic, changepoint) components.

    Notes
    -----
    The model uses a linear trend + periodic function + changepoints via XGBModule or LGBModule.
    The loss function used is Mean Squared Error (MSE). The trend component uses GBLinear.
    Changepoints allow the model to capture changes in the time series trend.
    """

    def __init__(
        self,
        nrounds=50,
        params={},
        module_type="XGBModule",
        linear_params={},
        changepoint_params={},
        estimate_uncertainty=True,
    ):
        self.nrounds = nrounds
        self.model_ = None
        self.losses_ = []
        self.module_type = module_type
        self.trend_type = "GBLinear"

        self.params = {"eta": 0.17, "max_depth": 3, "lambda": 1, "alpha": 8}
        self.params.update(params)

        self.linear_params = {"min_hess": 0.0, "lambd": 0.1, "lr": 0.9}
        self.linear_params.update(linear_params)

        self.changepoint_params = {
            "n_changepoints": 32,
            "eta": 0.9,
            "max_depth": 9,
            "lambda": 6.5,
            "alpha": 3.8,
            "cp_gap": 0.5,
            "cp_train_gap": 4,
        }
        self.changepoint_params.update(changepoint_params)
        self.estimate_uncertainty = estimate_uncertainty

    def fit(self, X, y=None):
        df = X.copy()
        df["y"] = y

        self.model_ = ForecastModule(
            df.shape[0],
            params=self.params,
            module_type=self.module_type,
            trend_type=self.trend_type,
            linear_params=self.linear_params,
            changepoint_params=self.changepoint_params,
        )
        self.model_.train()
        optimizer = torch.optim.Adam(self.model_.parameters(), lr=0.01)
        mse = torch.nn.MSELoss()

        for _ in range(self.nrounds):
            optimizer.zero_grad()
            preds = self.model_(df)
            loss = mse(preds.flatten(), torch.Tensor(df["y"].values).flatten())
            loss.backward(create_graph=True)
            self.losses_.append(loss.detach().item())
            self.model_.gb_step()

        if self.estimate_uncertainty:
            # Main idea: uncertainty = basic sigma2 + G * (test datetime - max train datetime)
            # G is fit using a train/validation on half the training data
            train_residuals = df["y"] - self.model_(df).detach().numpy().flatten()
            sigma2, G = _get_uncertainty_params(self, train_residuals, df)
            self.sigma2 = sigma2
            self.G = G
            # For better scaling, self.G was estimated in years
            self.max_train_years = (
                pd_datetime_to_seconds(df["ds"]).max() / 60 / 60 / 24 / 365
            )
            self.min_train_years = (
                pd_datetime_to_seconds(df["ds"]).min() / 60 / 60 / 24 / 365
            )

        self.model_.eval()
        return self

    def predict(self, X):
        check_is_fitted(self, "model_")
        df = X.copy()

        t, p, c = self.model_(df, components=True)
        forecast = t + p + c
        trend = t + c

        df["yhat"] = forecast.detach().numpy()
        df["trend"] = trend.detach().numpy()
        df["season"] = p.detach().numpy()

        if self.estimate_uncertainty:
            df["years"] = pd_datetime_to_seconds(df["ds"]) / 60 / 60 / 24 / 365
            df["high_h"] = df["years"] - self.max_train_years
            df["low_h"] = self.min_train_years - df["years"]
            df["h"] = df[["high_h", "low_h"]].max(axis=1)
            df.loc[df["h"] < 0, "h"] = 0
            df["var_est"] = self.sigma2 + self.G * df["h"]

        return_cols = [
            c for c in ["ds", "y", "yhat", "trend", "season", "var_est"] if c in df
        ]
        return df[return_cols].copy()


class ForecastModule(torch.nn.Module):
    """PyTorch module for time series forecasting.

    This module combines a linear trend component with a periodic function and changepoints
    learned through gradient boosting to model time series data. The trend is modeled using
    GBLinear layer, while the periodic patterns and changepoints are captured by either
    XGBoost or LightGBM.

    Parameters
    ----------
    n : int
        Number of samples in training data
    params : dict, optional
        Parameters passed to the gradient boosting model. Defaults to None.
    module_type : str, optional
        Type of gradient boosting module to use, either "XGBModule" or "LGBModule".
        Defaults to "XGBModule".
    trend_type : str, optional
        Type of trend model to use, either "PyTorch" or "GBLinear". Defaults to "GBLinear".
    linear_params : dict, optional
        Parameters passed to GBLinear trend model if trend_type="GBLinear". Defaults to {}.
    changepoint_params : dict, optional
        Parameters for changepoint detection and modeling. Defaults to:
        {
            "n_changepoints": 100,
            "gbmodule": "XGBModule",
            "cp_gap": 0.9,
            "cp_train_gap": 10
        }

    Attributes
    ----------
    trend : Union[torch.nn.Linear, GBLinear]
        Linear layer for modeling trend component, either PyTorch Linear or GBLinear
    bn : torch.nn.BatchNorm1d
        Batch normalization layer (only used with PyTorch trend)
    periodic_fn : XGBModule or LGBModule
        Gradient boosting module for modeling periodic patterns
    trend_fn : XGBModule or LGBModule
        Gradient boosting module for modeling changepoints
    initialized : bool
        Whether the model has been initialized with initial trend estimates
    trend_type : str
        Type of trend model being used

    Methods
    -------
    initialize(df)
        Initializes trend parameters using least squares regression
    forward(df, components=False)
        Forward pass combining trend, periodic, and changepoint components
    gb_step()
        Performs one gradient boosting step for all components
    """

    def __init__(
        self,
        n,
        params=None,
        module_type="XGBModule",
        trend_type="GBLinear",
        linear_params={},
        changepoint_params={},
    ):
        super(ForecastModule, self).__init__()
        assert trend_type in {"PyTorch", "GBLinear"}

        self.initialized = False
        self.trend_type = trend_type

        if trend_type == "PyTorch":
            self.trend = torch.nn.Linear(1, 1)
            self.bn = nn.LazyBatchNorm1d()
        if trend_type == "GBLinear":
            self.trend = GBLinear(1, 1, **linear_params)

        GBModule = loadModule(module_type)
        self.periodic_fn = GBModule(
            batch_size=n,
            input_dim=6,  # year, month, day, hour, minute, weekday
            output_dim=1,
            params=params,
        )

        # initialize changepoints components
        cp_params = {
            "n_changepoints": 100,
            "gbmodule": "XGBModule",
            "cp_gap": 0.9,
            "cp_train_gap": 10,
        }
        cp_params.update(changepoint_params)

        self.n_changepoints = cp_params.pop("n_changepoints")
        self.cp_gap = cp_params.pop("cp_gap")
        self.cp_module = cp_params.pop("gbmodule")
        self.cp_train_gap = cp_params.pop("cp_train_gap")

        self.trend_fn = loadModule(self.cp_module)(
            batch_size=max(1, self.n_changepoints),
            input_dim=1,
            output_dim=1,
            params=cp_params,
        )
        self.cp_train_count = 0
        self.use_cp = True if self.n_changepoints > 0 else False

    def _changepoint_initialize(self, df):
        ncp = self.n_changepoints if self.use_cp else 1
        self.cp_input = np.linspace(
            df["numeric_dt"].min(),
            df["numeric_dt"].min()
            + self.cp_gap * (df["numeric_dt"].max() - df["numeric_dt"].min()),
            ncp + 2,
        )[1:-1].reshape([-1, 1])

    def _gblinear_initialize(self, df):
        X = df.copy()
        X = (
            X[X["numeric_dt"].notnull() & X["y"].notnull()]
            .reset_index(drop=True)
            .copy()
        )

        X["intercept"] = 1
        ests = lstsq(X[["intercept", "numeric_dt"]], X[["y"]])[0]

        with torch.no_grad():
            self.trend.linear.weight.copy_(torch.Tensor(ests[1:, :]))
            self.trend.linear.bias.copy_(torch.Tensor(ests[0]))

        self.initialized = True

    def initialize(self, df):
        self._changepoint_initialize(df)
        if self.trend_type == "GBLinear":
            self._gblinear_initialize(df)
            return

        X = df.copy()
        X = (
            X[X["numeric_dt"].notnull() & X["y"].notnull()]
            .reset_index(drop=True)
            .copy()
        )

        X["intercept"] = 1
        X["numeric_dt"] = (X["numeric_dt"] - X["numeric_dt"].mean()) / X[
            "numeric_dt"
        ].std()
        ests = lstsq(X[["intercept", "numeric_dt"]], X[["y"]])[0]

        with torch.no_grad():
            self.trend.weight.copy_(torch.Tensor(ests[1:, :]))
            self.trend.bias.copy_(torch.Tensor(ests[0]))

        self.initialized = True

    def forward(self, df, components=False):
        df = df.copy()
        # Assume df formatted like prophet (columns 'ds' and 'y')
        df["year"] = df["ds"].dt.year
        df["month"] = df["ds"].dt.month
        df["day"] = df["ds"].dt.day
        df["hour"] = df["ds"].dt.hour
        df["minute"] = df["ds"].dt.minute
        df["weekday"] = df["ds"].dt.weekday
        df["numeric_dt"] = pd_datetime_to_seconds(df["ds"])
        if not self.initialized:
            self.initialize(df)

        X = np.array(
            df[["year", "month", "day", "hour", "minute", "weekday"]]
        )  # TODO just need to do this once

        if self.trend_type == "PyTorch":
            trend_component = self.trend(
                self.bn(
                    torch.Tensor(np.array(pd_datetime_to_seconds(df["ds"]))).reshape(
                        [-1, 1]
                    )
                )
            )
        if self.trend_type == "GBLinear":
            trend_component = self.trend(
                torch.Tensor(np.array(pd_datetime_to_seconds(df["ds"]))).reshape(
                    [-1, 1]
                )
            )

        gb_trend = self.forward_changepoints(df)

        forecast = trend_component + self.periodic_fn(X) + gb_trend
        if components:
            return trend_component, self.periodic_fn(X), gb_trend

        return forecast

    def forward_changepoints(self, df):
        slope_adjustments = self.trend_fn(self.cp_input)
        if not self.use_cp:
            return torch.zeros([df.shape[0], 1])
        changepoints = torch.concatenate(
            [torch.Tensor(self.cp_input), slope_adjustments], axis=1
        )
        cuml_adj = piecewise_linear_function(
            changepoints, torch.Tensor(np.array(df["numeric_dt"]))
        )
        return cuml_adj.reshape([-1, 1])

    def gb_step(self):
        if self.trend_type == "GBLinear":
            self.trend.gb_step()

        if not self.use_cp:
            self.periodic_fn.gb_step()
            return

        self.trend_fn.gb_step()
        self.cp_train_count += 1
        if self.cp_train_count > self.cp_train_gap:
            self.periodic_fn.gb_step()


def piecewise_linear_function(changepoints, timepoints):
    """
    Calculate a piecewise linear function based on changepoints evaluated at timepoints.

    For every tp in timepoints, compute:
    Sum_{(cp1, cp2) such that cp1 < tp} cp2 * (tp - cp1)

    Args:
        changepoints: torch.Tensor of shape [N, 2] where:
            - First column (cp1) contains the positions of changepoints
            - Second column (cp2) contains the slopes at those changepoints
        timepoints: torch.Tensor of shape [M] containing points to evaluate the function

    Returns:
        torch.Tensor of shape [M] containing the evaluated function at each timepoint
    """
    # Extract cp1 and cp2 from changepoints
    cp1 = changepoints[:, 0]  # Shape: [N]
    cp2 = changepoints[:, 1]  # Shape: [N]

    # Reshape for broadcasting
    # cp1: [N, 1], timepoints: [1, M] for broadcasting to [N, M]
    cp1_expanded = cp1.unsqueeze(1)  # Shape: [N, 1]
    timepoints_expanded = timepoints.unsqueeze(0)  # Shape: [1, M]

    # Create mask where cp1 < tp (shape: [N, M])
    mask = (cp1_expanded < timepoints_expanded).float()

    # Calculate (tp - cp1) for all pairs (shape: [N, M])
    time_diffs = (timepoints_expanded - cp1_expanded) * mask

    # Multiply by cp2 (shape: [N, M])
    cp2_expanded = cp2.unsqueeze(1)  # Shape: [N, 1]
    contributions = cp2_expanded * time_diffs

    # Sum over all changepoints (axis 0) to get final result for each timepoint
    result = torch.sum(contributions, dim=0)  # Shape: [M]

    return result


def _get_uncertainty_params(m, train_residuals, train):
    utrain_cutoff = train["ds"].quantile(0.5)
    utrain = (
        train[train["ds"] <= utrain_cutoff]
        .sort_values("ds")
        .reset_index(drop=True)
        .copy()
    )
    utest = (
        train[train["ds"] > utrain_cutoff]
        .sort_values("ds")
        .reset_index(drop=True)
        .copy()
    )

    um = Forecast(
        nrounds=m.nrounds,
        params=m.params,
        module_type=m.module_type,
        linear_params=m.linear_params,
        changepoint_params=m.changepoint_params,
        estimate_uncertainty=False,
    )
    um.fit(utrain, utrain["y"])
    utrain["gbnet_pred"] = um.predict(utrain)["yhat"].copy()
    utest["gbnet_pred"] = um.predict(utest)["yhat"].copy()

    utrain["residual"] = utrain["y"] - utrain["gbnet_pred"]
    utest["residual"] = utest["y"] - utest["gbnet_pred"]

    try:
        G, info = _mle_G(
            utest["residual"],
            pd_datetime_to_seconds(utest["ds"]) / 60 / 60 / 24 / 365
            - pd_datetime_to_seconds(utrain["ds"]).max() / 60 / 60 / 24 / 365,
            (utrain["residual"] ** 2).sum() / (utrain.shape[0] - 1),
        )
    except (
        ValueError
    ):  # TODO maybe we estimate sigma2 and G simultaneously? to avoid this issue?
        G = 0.0

    sigma2 = (train_residuals**2).sum() / (train.shape[0] - 1)

    return sigma2, G


def _mle_G(r, h, S2, bracket_pad=1.0):
    """
    MLE for G in Var(r_i)=S2 + G*h_i with known S2.
    Returns (G_hat, diagnostics)
    """
    r = np.asarray(r, dtype=float)
    h = np.asarray(h, dtype=float)
    assert (h >= 0).all(), "h_i must be non-negative"

    def score(G):
        denom = S2 + G * h
        output = np.sum(h * (S2 + G * h - r**2) / denom**2)
        return output

    # Find a bracket [a,b] where score(a)>0, score(b)<0
    a = 0.0
    b = max(np.max((r**2 - S2) / (h + 1e-12)), 1.0) + bracket_pad
    sol = root_scalar(score, bracket=(a, b), method="brentq")
    return sol.root, sol



import unittest
import numpy as np
import pandas as pd
from gbnet.models import ordinal_regression


class TestOrdinalRegression(unittest.TestCase):
    def setUp(self):
        # Generate synthetic ordinal data
        num_samples = 200
        num_features = 3
        self.num_classes = 4

        # Generate random features
        self.X = np.random.randn(num_samples, num_features)

        # Generate ordinal targets (0 to num_classes-1)
        self.y = pd.Series(np.random.randint(0, self.num_classes, num_samples))

    def test_fit_predict(self):
        """Test that the estimator can fit and predict without errors."""
        # Split data into train/test
        train_idx = range(150)
        test_idx = range(150, 200)

        X_train = self.X[train_idx]
        y_train = self.y[train_idx]
        X_test = self.X[test_idx]

        # Initialize and fit estimator
        model = ordinal_regression.GBOrd(num_classes=self.num_classes, nrounds=10)
        model.fit(X_train, y_train)

        # Make predictions
        predictions = model.predict(X_test)

        # Check predictions have correct length
        self.assertEqual(len(predictions), len(X_test))

        # Check predictions are not NaN
        self.assertFalse(np.isnan(predictions).any())

        # Check predictions are numerical
        self.assertTrue(np.issubdtype(predictions.dtype, np.number))

    def test_loss_decreases(self):
        """Test that the loss decreases during training"""
        model = ordinal_regression.GBOrd(num_classes=self.num_classes, nrounds=10)
        model.fit(self.X[:50], self.y[:50])

        # Check losses are recorded
        self.assertTrue(len(model.losses_) > 0)

        # Check loss generally decreases
        self.assertLess(np.mean(model.losses_[-3:]), np.mean(model.losses_[:3]))

    def test_invalid_num_classes(self):
        """Test that model raises error if num_classes doesn't match data"""
        with self.assertRaises(AssertionError):
            model = ordinal_regression.GBOrd(
                num_classes=10
            )  # More classes than in data
            model.fit(self.X, self.y)

    def test_fit_predict_xgb(self):
        """Test that the estimator can fit and predict without errors."""
        # Split data into train/test
        train_idx = range(150)
        test_idx = range(150, 200)

        X_train = self.X[train_idx]
        y_train = self.y[train_idx]
        X_test = self.X[test_idx]

        # Initialize and fit estimator
        model = ordinal_regression.GBOrd(
            num_classes=self.num_classes, nrounds=10, module_type="XGBModule"
        )
        model.fit(X_train, y_train)

        # Make predictions
        predictions = model.predict(X_test)

        # Check predictions have correct length
        self.assertEqual(len(predictions), len(X_test))

        # Check predictions are not NaN
        self.assertFalse(np.isnan(predictions).any())

        # Check predictions are numerical
        self.assertTrue(np.issubdtype(predictions.dtype, np.number))

    def test_loss_decreases_xgb(self):
        """Test that the loss decreases during training"""
        model = ordinal_regression.GBOrd(
            num_classes=self.num_classes, nrounds=10, module_type="XGBModule"
        )
        model.fit(self.X[:50], self.y[:50])

        # Check losses are recorded
        self.assertTrue(len(model.losses_) > 0)

        # Check loss generally decreases
        self.assertLess(np.mean(model.losses_[-3:]), np.mean(model.losses_[:3]))

    def test_invalid_num_classes_xgb(self):
        """Test that model raises error if num_classes doesn't match data"""
        with self.assertRaises(AssertionError):
            model = ordinal_regression.GBOrd(
                num_classes=10, module_type="XGBModule"
            )  # More classes than in data
            model.fit(self.X, self.y)



import numpy as np
from sklearn.base import BaseEstimator, ClassifierMixin
from sklearn.utils.validation import check_is_fitted
import torch
from torch import nn


def loadModule(module):
    assert module in {"XGBModule", "LGBModule"}
    if module == "XGBModule":
        from gbnet import xgbmodule

        return xgbmodule.XGBModule
    if module == "LGBModule":
        from gbnet import lgbmodule

        return lgbmodule.LGBModule


class GBOrd(BaseEstimator, ClassifierMixin):
    """Gradient Boosting Ordinal Regression model.

    This model combines gradient boosting with ordinal regression to predict ordered
    categorical outcomes. It uses either XGBoost or LightGBM as the underlying boosting
    engine wrapped in a PyTorch module.

    Parameters
    ----------
    num_classes : int
        Number of ordinal classes to predict
    nrounds : int, optional
        Number of boosting rounds. Defaults to 500 for XGBModule and 1000 for LGBModule.
    params : dict, optional
        Additional parameters passed to the gradient boosting model.
    module_type : str, optional
        Type of gradient boosting module to use, either "XGBModule" or "LGBModule".
        Defaults to "LGBModule".
    min_hess : float, optional
        Minimum hessian value for numerical stability. Defaults to 0.0.

    Attributes
    ----------
    model_ : XGBModule or LGBModule
        Trained gradient boosting module. Set after fitting.
    losses_ : list
        List of loss values recorded at each training iteration.
    min_targets : int
        Minimum value in training targets, used for label normalization.

    Methods
    -------
    fit(X, y)
        Trains the model using input features X and ordinal targets y.
    predict(X)
        Predicts ordinal class labels for input features X.

    Notes
    -----
    The model uses an ordinal logistic loss function to handle ordered categorical outcomes.
    The gradient boosting model learns a single score which is transformed into class
    probabilities via learned thresholds.
    """

    def __init__(
        self,
        num_classes,
        nrounds=None,
        params=None,
        module_type="LGBModule",
        min_hess=0.0,
    ):
        if params is None:
            params = {}
        if nrounds is None:
            if module_type == "XGBModule":
                nrounds = 500
            if module_type == "LGBModule":
                nrounds = 1000
        self.nrounds = nrounds
        self.params = params
        self.model_ = None
        self.losses_ = []
        self.module_type = module_type
        self.Module = loadModule(module_type)
        self.loss_fn = OrdinalLogisticLoss(num_classes=num_classes)
        self.num_classes = num_classes
        self.min_hess = min_hess

    def fit(self, X, y=None):
        self.min_targets = min(y)
        targets = torch.Tensor(y.values).flatten()
        targets = targets.long()
        targets = targets - self.min_targets
        assert len(np.unique(y)) == self.num_classes

        self.model_ = self.Module(
            X.shape[0], X.shape[1], 1, params=self.params, min_hess=self.min_hess
        )
        self.model_.train()

        optimizer = torch.optim.Adam(
            list(self.model_.parameters()) + list(self.loss_fn.parameters()), lr=0.01
        )

        for _ in range(self.nrounds):
            optimizer.zero_grad()
            preds = self.model_(X).flatten()
            loss = self.loss_fn(preds, targets)
            loss.backward(create_graph=True)
            self.losses_.append(loss.detach().item())
            self.model_.gb_step()
            optimizer.step()

        self.model_.eval()
        return self

    def score(self, X, y):
        """
        Return the negative log likelihood score for input X and targets y.

        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            Input features.
        y : array-like of shape (n_samples,)
            Target values.

        Returns
        -------
        float
            Negative log likelihood score. Lower values indicate better fit.
        """
        check_is_fitted(self, "model_")
        targets = torch.Tensor(y.values).flatten().long()
        targets = targets - torch.min(targets)
        logits = self.model_(X).flatten()
        neg_log_likelihood = self.loss_fn(logits, targets)
        return neg_log_likelihood.detach().item()

    def predict_proba(self, X):
        """
        Predict class probabilities for input X.

        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            Input features.

        Returns
        -------
        array-like of shape (n_samples, n_classes)
            Predicted class probabilities.
        """
        check_is_fitted(self, "model_")
        logits = self.model_(X).flatten()
        probs = self.loss_fn.get_pred_probs(logits).detach().numpy()
        return probs

    def predict(self, X, return_logits=True):
        """
        Predict continuous output for input X.
        """
        check_is_fitted(self, "model_")
        if return_logits:
            preds = self.model_(X).detach().numpy()
            return preds.flatten()
        else:
            preds = self.predict_proba(X).argmax(axis=1) + self.min_targets
            return preds


class OrdinalLogisticLoss(nn.Module):
    def __init__(self, num_classes):
        super().__init__()
        self.num_classes = num_classes

        self.breakpoints = nn.Parameter(
            torch.arange(num_classes - 1, dtype=torch.float32)
            - (num_classes - 2.0) / 2.0
        )

    def _compute_probabilities(self, logits):
        """
        Compute class probabilities
        """
        if logits.dim() == 2:
            logits = logits.squeeze(1)

        # Compute cumulative probabilities
        cum_probs = torch.sigmoid(self.breakpoints.unsqueeze(0) - logits.unsqueeze(1))

        eps = 1e-8
        probs = torch.diff(
            cum_probs,
            prepend=torch.zeros_like(cum_probs[:, :1]),
            append=torch.ones_like(cum_probs[:, :1]),
        ).clamp(min=eps)

        return probs

    def forward(self, logits, targets):
        """
        Compute loss more efficiently but maintaining numerical equivalence.
        """
        targets = targets.flatten()
        probs = self._compute_probabilities(logits)

        # More efficient than one-hot but numerically equivalent
        target_probs = probs[torch.arange(probs.size(0)), targets]
        loss = -torch.log(target_probs).mean()

        return loss

    def get_pred_probs(self, logits):
        """
        Predict most likely class.
        """
        return self._compute_probabilities(logits)



import numpy as np
import torch
from catboost import CatBoostRegressor
from gbnet.experimental import catmodule


def test_basic_catboost_example():
    np.random.seed(100)
    n = 10
    xs = np.random.random([n, 4])
    ys = np.random.random([n, 2])
    niters = 10

    model = CatBoostRegressor(
        iterations=niters,
        learning_rate=0.03,
        boost_from_average=False,
        loss_function="MultiRMSE",
    )
    model.fit(xs, ys)
    pred = model.predict(xs)

    cbnet = catmodule.CatModule(
        xs.shape[0],
        xs.shape[1],
        ys.shape[1],
        params={
            "learning_rate": 0.03,
        },
    )
    cbmse = torch.nn.MSELoss()
    for i in range(niters):
        cbnet.zero_grad()
        cbpred = cbnet(xs)

        loss = cbmse(cbpred, torch.Tensor(ys))
        loss.backward(create_graph=True)

        cbnet.gb_step(xs)
    cbnet.eval()

    assert np.isclose(pred, cbnet(xs).detach().numpy()).all()



import unittest
from unittest.mock import MagicMock, patch

from datetime import datetime, timedelta

import numpy as np
import pandas as pd
from pandas.testing import assert_series_equal, assert_frame_equal
from gbnet.models import forecasting
import torch


class TestForecast(unittest.TestCase):
    def setUp(self):
        # Generate synthetic data
        num_samples = 200
        self.dates = [datetime.now() + timedelta(days=i) for i in range(num_samples)]
        self.y = np.sin(np.linspace(0, 20, num_samples))  # Synthetic target variable
        self.df = pd.DataFrame({"ds": self.dates, "y": self.y})

    def test_fit_predict(self):
        """
        Test that the estimator can fit and predict without errors using GBLinear trend.
        """
        # Split data into training and testing sets
        train_df = self.df.iloc[:150]
        test_df = self.df.iloc[150:]

        # Initialize and fit the estimator
        estimator = forecasting.Forecast(nrounds=10)
        estimator.fit(train_df[["ds"]], train_df["y"])

        # Make predictions
        predictions = estimator.predict(test_df[["ds"]])

        # Check that predictions have the correct length
        self.assertEqual(
            predictions.shape[0], len(test_df), "Prediction length mismatch."
        )

        # Check that predictions are not NaN
        self.assertFalse(
            np.isnan(predictions).any().any(), "Predictions contain NaN values."
        )

    def test_fit_predict_lgbm(self):
        """
        Test that the estimator can fit and predict without errors using LGBModule and GBLinear trend.
        """
        # Split data into training and testing sets
        train_df = self.df.iloc[:150]
        test_df = self.df.iloc[150:]

        # Initialize and fit the estimator
        estimator = forecasting.Forecast(nrounds=10, module_type="LGBModule")
        estimator.fit(train_df[["ds"]], train_df["y"])

        # Make predictions
        predictions = estimator.predict(test_df[["ds"]])

        # Check that predictions have the correct length
        self.assertEqual(
            predictions.shape[0], len(test_df), "Prediction length mismatch."
        )

        # Check that predictions are not NaN
        self.assertFalse(
            np.isnan(predictions).any().any(), "Predictions contain NaN values."
        )

    def test_loss_decrease(self):
        """
        Test that the training loss decreases over epochs with GBLinear trend.
        """
        # Use only the training data
        train_df = self.df.iloc[:150]

        # Initialize and fit the estimator
        estimator = forecasting.Forecast(nrounds=10)
        estimator.fit(train_df[["ds"]], train_df["y"])

        # Get the recorded losses
        losses = estimator.losses_

        # Check that losses have been recorded
        self.assertTrue(len(losses) > 0, "No losses recorded during training.")

        # Check that the loss decreases over time
        initial_loss = losses[0]
        final_loss = losses[-1]
        self.assertLess(
            final_loss, initial_loss, "Final loss is not less than initial loss."
        )

        # Check that the losses are decreasing monotonically (not required but useful)
        self.assertTrue(
            all(x >= y for x, y in zip(losses[:10], losses[1:11])),
            "Losses are not monotonically decreasing.",
        )

    def test_loss_decrease_lgbm(self):
        """
        Test that the training loss decreases over epochs with LGBModule and GBLinear trend.
        """
        # Use only the training data
        train_df = self.df.iloc[:150]

        # Initialize and fit the estimator
        estimator = forecasting.Forecast(nrounds=10, module_type="LGBModule")
        estimator.fit(train_df[["ds"]], train_df["y"])

        # Get the recorded losses
        losses = estimator.losses_

        # Check that losses have been recorded
        self.assertTrue(len(losses) > 0, "No losses recorded during training.")

        # Check that the loss decreases over time
        initial_loss = losses[0]
        final_loss = losses[-1]
        self.assertLess(
            final_loss, initial_loss, "Final loss is not less than initial loss."
        )

        # Check that the losses are decreasing monotonically (not required but useful)
        self.assertTrue(
            all(x >= y for x, y in zip(losses[:10], losses[1:11])),
            "Losses are not monotonically decreasing.",
        )

    def test_changepoint_initialization(self):
        """
        Test that changepoints are properly initialized with the correct number and spacing.
        """
        # Initialize estimator with changepoint parameters
        estimator = forecasting.Forecast(
            nrounds=10,
            changepoint_params={
                "n_changepoints": 5,
                "cp_gap": 0.8,
                "cp_train_gap": 2,
            },
        )

        # Fit the model
        estimator.fit(self.df[["ds"]], self.df["y"])

        # Check that changepoints were initialized
        self.assertTrue(
            hasattr(estimator.model_, "cp_input"), "Changepoints not initialized"
        )
        self.assertEqual(
            len(estimator.model_.cp_input), 5, "Incorrect number of changepoints"
        )

        # Check changepoint spacing
        cp_times = estimator.model_.cp_input.flatten()
        expected_min = self.df["ds"].min().timestamp()
        expected_max = self.df["ds"].min().timestamp() + 0.8 * (
            self.df["ds"].max().timestamp() - self.df["ds"].min().timestamp()
        )
        self.assertGreaterEqual(
            cp_times[0], expected_min, "First changepoint too early"
        )
        self.assertLessEqual(cp_times[-1], expected_max, "Last changepoint too late")

    def test_changepoint_prediction(self):
        """
        Test that predictions with changepoints work correctly and return expected components.
        """
        # Initialize estimator with changepoints
        estimator = forecasting.Forecast(
            nrounds=10, changepoint_params={"n_changepoints": 3}
        )

        # Fit the model
        estimator.fit(self.df[["ds"]], self.df["y"])

        # Get predictions with components
        pred_df = estimator.predict(self.df[["ds"]])

        # Check component shapes and nans
        self.assertEqual(self.df.shape[0], pred_df.shape[0])
        self.assertEqual(pd.isnull(self.df).sum().sum(), 0)

    def test_changepoint_impact(self):
        """
        Test that changepoints have a meaningful impact on predictions.
        """
        # Create a dataset with a clear changepoint
        dates = pd.date_range(start="2020-01-01", periods=200, freq="D")
        y = np.concatenate(
            [
                np.linspace(0, 1, 100),  # First trend
                np.linspace(1, 2, 100),  # Second trend
            ]
        )
        df = pd.DataFrame({"ds": dates, "y": y})

        # Fit model with and without changepoints
        model_with_cp = forecasting.Forecast(
            nrounds=10, changepoint_params={"n_changepoints": 3}
        )
        model_without_cp = forecasting.Forecast(
            nrounds=10, changepoint_params={"n_changepoints": 0}
        )

        model_with_cp.fit(df[["ds"]], df["y"])
        model_without_cp.fit(df[["ds"]], df["y"])

        # Get predictions
        preds_with_cp = model_with_cp.predict(df[["ds"]])
        preds_without_cp = model_without_cp.predict(df[["ds"]])

        # Calculate MSE for both models
        mse_with_cp = np.mean((preds_with_cp["yhat"] - df["y"]) ** 2)
        mse_without_cp = np.mean((preds_without_cp["yhat"] - df["y"]) ** 2)

        # Check that changepoints improve the fit
        self.assertLess(
            mse_with_cp, mse_without_cp, "Changepoints did not improve model fit"
        )

    def test_changepoint_parameters(self):
        """
        Test that different changepoint parameters affect the model behavior.
        """
        # Test different numbers of changepoints
        for n_cp in [0, 1, 5, 10]:
            estimator = forecasting.Forecast(
                nrounds=10, changepoint_params={"n_changepoints": n_cp}
            )
            estimator.fit(self.df[["ds"]], self.df["y"])

            # Check that changepoints were initialized correctly
            if n_cp > 0:
                self.assertEqual(
                    len(estimator.model_.cp_input),
                    n_cp,
                    f"Incorrect number of changepoints for n_cp={n_cp}",
                )
            else:
                self.assertFalse(
                    estimator.model_.use_cp, "Changepoints should be disabled"
                )

            # Check that predictions work
            preds = estimator.predict(self.df[["ds"]])
            self.assertEqual(len(preds), len(self.df), "Prediction length mismatch")
            self.assertFalse(
                np.isnan(preds["yhat"]).any(), "Predictions contain NaN values"
            )

    def test_piecewise_linear_function(self):
        """
        Test the piecewise_linear_function implementation with various scenarios.
        """
        # Test case 1: Single changepoint
        changepoints = torch.tensor(
            [[1.0, 2.0]], dtype=torch.float32
        )  # [position, slope]
        timepoints = torch.tensor([0.0, 0.5, 1.0, 1.5, 2.0], dtype=torch.float32)
        result = forecasting.piecewise_linear_function(changepoints, timepoints)

        # For t < 1.0, result should be 0
        # For t >= 1.0, result should be 2.0 * (t - 1.0)
        expected = torch.tensor([0.0, 0.0, 0.0, 1.0, 2.0], dtype=torch.float32)
        self.assertTrue(
            torch.allclose(result, expected), "Single changepoint test failed"
        )

        # Test case 2: Multiple changepoints
        changepoints = torch.tensor(
            [
                [1.0, 1.0],  # First changepoint at t=1 with slope 1
                [2.0, 2.0],  # Second changepoint at t=2 with slope 2
            ],
            dtype=torch.float32,
        )
        timepoints = torch.tensor([0.0, 1.0, 1.5, 2.0, 2.5, 3.0], dtype=torch.float32)
        result = forecasting.piecewise_linear_function(changepoints, timepoints)

        # Expected results:
        # t < 1.0: 0 (no changepoints before)
        # 1.0 <= t < 2.0: 1.0 * (t - 1.0) (only first changepoint contributes)
        # t >= 2.0: 1.0 * (t - 1.0) + 2.0 * (t - 2.0) (both changepoints contribute)
        expected = torch.tensor([0.0, 0.0, 0.5, 1.0, 2.5, 4.0], dtype=torch.float32)
        self.assertTrue(
            torch.allclose(result, expected), "Multiple changepoints test failed"
        )

        # Test case 3: No changepoints
        changepoints = torch.tensor([], dtype=torch.float32).reshape(0, 2)
        timepoints = torch.tensor([0.0, 1.0, 2.0], dtype=torch.float32)
        result = forecasting.piecewise_linear_function(changepoints, timepoints)
        expected = torch.zeros_like(timepoints)
        self.assertTrue(torch.allclose(result, expected), "No changepoints test failed")

        # Test case 4: Changepoints with negative slopes
        changepoints = torch.tensor(
            [
                [1.0, -1.0],  # First changepoint at t=1 with slope -1
                [2.0, -2.0],  # Second changepoint at t=2 with slope -2
            ],
            dtype=torch.float32,
        )
        timepoints = torch.tensor([0.0, 1.0, 1.5, 2.0, 2.5, 3.0], dtype=torch.float32)
        result = forecasting.piecewise_linear_function(changepoints, timepoints)

        # Expected results:
        # t < 1.0: 0
        # 1.0 <= t < 2.0: -1.0 * (t - 1.0)
        # t >= 2.0: -1.0 * (t - 1.0) - 2.0 * (t - 2.0)
        expected = torch.tensor([0.0, 0.0, -0.5, -1.0, -2.5, -4.0], dtype=torch.float32)
        self.assertTrue(torch.allclose(result, expected), "Negative slopes test failed")


class TestMleG(unittest.TestCase):
    def test_mle_G_when_true_G_is_zero_perfect_data(self):
        """Test _mle_G when true G is zero (r_i^2 = S2)."""
        S2 = 1.0
        h = np.array([0.0, 1.0, 2.0, 3.0, 4.0, 5.0])  # Needs some non-zero h
        r_sq = np.full_like(h, S2)
        r = np.sqrt(r_sq)

        G_estimated, _ = forecasting._mle_G(
            r, h, S2, bracket_pad=0.01
        )  # Small pad for precision
        self.assertAlmostEqual(G_estimated, 0.0, places=5)

    def test_mle_G_when_true_G_is_zero_h_all_zeros(self):
        """Test _mle_G when h is all zeros, G should be 0."""
        S2 = 1.0
        h = np.zeros(10)
        r = np.full(10, np.sqrt(S2))  # r^2 = S2 exactly

        # If all h are zero, score function is identically zero.
        # root_scalar might return 'a' (0.0) from bracket (a,b)
        G_estimated, _ = forecasting._mle_G(r, h, S2, bracket_pad=0.1)
        self.assertAlmostEqual(G_estimated, 0.0, places=5)

    def test_mle_G_positive_G_perfect_fit(self):
        """Test _mle_G with a known positive G and no noise."""
        S2 = 1.0
        G_true = 0.5
        h = np.array([0.0, 1.0, 2.0, 3.0, 4.0, 5.0])
        # r_i^2 = S2 + G_true * h_i
        r = np.sqrt(S2 + G_true * h)

        G_estimated, _ = forecasting._mle_G(r, h, S2, bracket_pad=0.01)
        self.assertAlmostEqual(G_estimated, G_true, places=5)

    def test_mle_G_positive_G_with_noise(self):
        """Test _mle_G with positive G and noisy r."""
        S2 = 1.0
        G_true = 0.5
        N = 200  # Larger N for more stable estimate with noise
        h = np.random.rand(N) * 5
        h[0] = 0.0  # Ensure some h can be zero

        # r_i ~ N(0, variance = S2 + G_true * h_i)
        variances = S2 + G_true * h
        # Ensure variances are positive
        self.assertTrue(np.all(variances > 0.00001))  # Avoid zero variance

        r = np.array([np.random.normal(0, np.sqrt(v)) for v in variances])

        G_estimated, _ = forecasting._mle_G(r, h, S2, bracket_pad=1.0)
        # With noise, G_estimated won't be exact. Check if it's reasonably close.
        # This is a statistical estimate, so large deviation is possible for small N.
        self.assertAlmostEqual(G_estimated, G_true, delta=0.35)  # Allow some deviation

    def test_mle_G_h_must_be_non_negative_assertion(self):
        """Test that _mle_G asserts h_i >= 0."""
        S2 = 1.0
        r = np.array([1.0, 1.0])
        h_negative = np.array([1.0, -1.0])
        with self.assertRaisesRegex(AssertionError, "h_i must be non-negative"):
            forecasting._mle_G(r, h_negative, S2)


class TestGetUncertaintyParams(unittest.TestCase):
    @patch("gbnet.models.forecasting.Forecast")  # Patch Forecast where it's looked up
    @patch("gbnet.models.forecasting._mle_G")  # Patch _mle_G where it's looked up
    def test_get_uncertainty_params_main_logic(self, mock_mle_G, MockForecastClass):
        """Test the main workflow of _get_uncertainty_params."""
        # 1. Mock for 'm', the main model instance passed to the function
        mock_m_outer = MagicMock()
        mock_m_outer.nrounds = 10
        mock_m_outer.params = {"key_param": "val_param"}
        mock_m_outer.module_type = "XGBModule"
        mock_m_outer.linear_params = {"lin_key": "lin_val"}
        mock_m_outer.changepoint_params = {"cp_key": "cp_val"}

        # 2. train_residuals (from outer model) and train_df
        N = 40
        self.assertTrue(N >= 4, "Need N>=4 for split and variance calcs with N-1 denom")
        original_train_residuals = np.random.normal(0, 1.5, N)
        expected_final_sigma2 = np.sum(original_train_residuals**2) / (N - 1)

        base_time = pd.Timestamp("2023-01-01")
        train_ds = pd.Series([base_time + pd.Timedelta(days=i) for i in range(N)])
        train_y = np.arange(N, dtype=float) * 0.5
        train_df = pd.DataFrame({"ds": train_ds, "y": train_y})

        # 3. Configure MockForecastClass for the internal 'um' model
        mock_um_instance = MockForecastClass.return_value

        # Determine utrain and utest splits as done in the actual function
        utrain_cutoff_ds = train_df["ds"].quantile(0.5)  # Inclusive for '<='
        utrain_df_internal = (
            train_df[train_df["ds"] <= utrain_cutoff_ds]
            .sort_values("ds")
            .reset_index(drop=True)
        )
        utest_df_internal = (
            train_df[train_df["ds"] > utrain_cutoff_ds]
            .sort_values("ds")
            .reset_index(drop=True)
        )

        N_utrain = utrain_df_internal.shape[0]
        N_utest = utest_df_internal.shape[0]
        self.assertTrue(
            N_utrain >= 2 and N_utest >= 1, "Splits resulted in too small DFs"
        )

        target_S2_for_mle_G = 2.0
        # utrain_residuals = y_utrain - yhat_utrain. We want var(utrain_residuals) = target_S2_for_mle_G
        # (sum of squares) / (N_utrain - 1) = target_S2_for_mle_G
        # If residuals are const c_res: c_res^2 * N_utrain / (N_utrain-1) approx target_S2_for_mle_G
        # More directly: sum(c_res^2) = target_S2_for_mle_G * (N_utrain - 1)
        # c_res = sqrt(target_S2_for_mle_G * (N_utrain - 1) / N_utrain)
        utrain_residual_val = np.sqrt(target_S2_for_mle_G * (N_utrain - 1) / N_utrain)

        yhat_utrain = utrain_df_internal["y"].values - utrain_residual_val
        utest_residuals_for_mle = np.random.normal(0, 1, N_utest)
        yhat_utest = utest_df_internal["y"].values - utest_residuals_for_mle

        # Configure predict method of mock_um_instance
        # It's called twice: first with utrain, then with utest.
        # Using a list of return values for side_effect is often cleaner.
        mock_um_instance.predict.side_effect = [
            pd.DataFrame({"yhat": yhat_utrain}),  # For utrain call
            pd.DataFrame({"yhat": yhat_utest}),  # For utest call
        ]
        mock_um_instance.fit.return_value = None

        # 4. Configure mock_mle_G
        expected_G_from_mle = 0.8
        mock_mle_G.return_value = (expected_G_from_mle, "mock_mle_diagnostics")

        # 5. Call the function under test
        returned_sigma2, returned_G = forecasting._get_uncertainty_params(
            mock_m_outer, original_train_residuals, train_df
        )

        # 6. Assertions
        self.assertAlmostEqual(returned_sigma2, expected_final_sigma2, places=5)
        self.assertEqual(returned_G, expected_G_from_mle)

        MockForecastClass.assert_called_once_with(
            nrounds=mock_m_outer.nrounds,
            params=mock_m_outer.params,
            module_type=mock_m_outer.module_type,
            linear_params=mock_m_outer.linear_params,
            changepoint_params=mock_m_outer.changepoint_params,
            estimate_uncertainty=False,
        )

        # Verify um.fit call arguments
        fit_call = mock_um_instance.fit.call_args
        self.assertIsNotNone(fit_call, "um.fit was not called")
        called_fit_df, called_fit_y = fit_call[0]
        assert_frame_equal(
            called_fit_df[["ds", "y"]], utrain_df_internal, check_dtype=False
        )
        assert_series_equal(called_fit_y, utrain_df_internal["y"], check_dtype=False)

        # Verify mock_mle_G was called correctly
        secs_in_year = 60 * 60 * 24 * 365.0
        expected_h_utrain_max_time_years = (
            forecasting.pd_datetime_to_seconds(utrain_df_internal["ds"]).max()
            / secs_in_year
        )
        expected_h_utest_ds_years = (
            forecasting.pd_datetime_to_seconds(utest_df_internal["ds"]) / secs_in_year
        )
        expected_h_values = (
            expected_h_utest_ds_years.values - expected_h_utrain_max_time_years
        )

        mle_G_call_args_list = mock_mle_G.call_args_list
        self.assertEqual(len(mle_G_call_args_list), 1, "_mle_G should be called once")
        called_r_for_mle, called_h_for_mle, called_S2_for_mle = mle_G_call_args_list[0][
            0
        ]

        np.testing.assert_array_almost_equal(
            called_r_for_mle, utest_residuals_for_mle, decimal=5
        )
        np.testing.assert_array_almost_equal(
            called_h_for_mle, expected_h_values, decimal=5
        )
        self.assertAlmostEqual(called_S2_for_mle, target_S2_for_mle_G, places=5)

    @patch("gbnet.models.forecasting.Forecast")
    @patch("gbnet.models.forecasting._mle_G")
    def test_get_uncertainty_params_mle_G_raises_ValueError(
        self, mock_mle_G, MockForecastClass
    ):
        """Test _get_uncertainty_params when _mle_G raises ValueError."""
        mock_m_outer = MagicMock()
        mock_m_outer.nrounds = 5
        mock_m_outer.params = {}
        mock_m_outer.module_type = "XGBModule"
        mock_m_outer.linear_params = {}
        mock_m_outer.changepoint_params = {}

        N = 20
        original_train_residuals = np.random.randn(N)
        expected_final_sigma2 = np.sum(original_train_residuals**2) / (N - 1)

        base_time = pd.Timestamp("2024-01-01")
        train_ds = pd.Series([base_time + pd.Timedelta(hours=i * 12) for i in range(N)])
        train_y = np.sin(np.linspace(0, 4 * np.pi, N))
        train_df = pd.DataFrame({"ds": train_ds, "y": train_y})

        mock_um_instance = MockForecastClass.return_value

        # Mock predict to return yhats that allow residuals to be calculated
        def predict_side_effect_for_value_error(df_arg):
            return pd.DataFrame({"yhat": np.zeros(df_arg.shape[0])})

        mock_um_instance.predict.side_effect = predict_side_effect_for_value_error
        mock_um_instance.fit.return_value = None

        mock_mle_G.side_effect = ValueError("MLE failed for testing")

        returned_sigma2, returned_G = forecasting._get_uncertainty_params(
            mock_m_outer, original_train_residuals, train_df
        )

        self.assertAlmostEqual(returned_sigma2, expected_final_sigma2, places=5)
        self.assertEqual(
            returned_G, 0.0, "G should default to 0.0 on ValueError from _mle_G"
        )
        mock_mle_G.assert_called_once()



from unittest import mock, TestCase

import numpy as np
import torch
import xgboost as xgb
from gbnet import xgbmodule as xgm


def test_basic_loss():
    gbm = xgm.XGBModule(5, 3, 1)
    floss = torch.nn.MSELoss()

    gbm.zero_grad()
    np.random.seed(11010)
    input_dmatrix = xgb.DMatrix(np.random.random([5, 3]))
    preds = gbm(input_dmatrix)
    loss = floss(preds.flatten(), torch.Tensor(np.array([1, 2, 3, 4, 5])).flatten())

    loss.backward(create_graph=True)

    with (
        mock.patch("gbnet.xgbmodule.XGBObj", side_effect=xgm.XGBObj) as m_obj,
        mock.patch.object(gbm.bst, "boost", side_effect=gbm.bst.boost) as m_boost,
    ):
        gbm.gb_step()

    assert np.all(
        np.isclose(
            m_obj.call_args_list[-1].args[0].detach().numpy(),
            np.array([-2, -4, -6, -8, -10]).reshape([-1, 1]),
        )
    )
    assert np.all(
        np.isclose(
            m_obj.call_args_list[-1].args[1].detach().numpy(),
            np.array([2, 2, 2, 2, 2]).reshape([-1, 1]),
        )
    )

    m_boost.assert_called_once()


def test_XGBObj():
    np.random.seed(10101)
    grad = torch.tensor(np.random.random([20, 10]))
    hess = torch.tensor(np.random.random([20, 10]))

    obj = xgm.XGBObj(grad, hess)

    with mock.patch("xgboost.__version__", new="2.0.0"):
        ograd, ohess = obj(grad, hess)
        assert np.all(np.isclose(ograd, grad.detach().numpy().reshape([-1, 1])))
        assert np.all(np.isclose(ohess, hess.detach().numpy().reshape([-1, 1])))
    with mock.patch("xgboost.__version__", new="2.1.0"):
        ograd, ohess = obj(grad, hess)
        assert np.all(np.isclose(ograd, grad.detach().numpy()))
        assert np.all(np.isclose(ohess, hess.detach().numpy()))


class TestInputChecking(TestCase):
    def test_input_is_dmatrix_training_true_dtrain_none(self):
        """Test with xgb.DMatrix input, training mode True, and dtrain is None."""
        module = xgm.XGBModule(10, 5, 3)
        data = np.random.rand(10, 5)
        dmatrix = xgb.DMatrix(data)
        result = module._input_checking_setting(dmatrix)
        self.assertIs(result, module.dtrain)
        self.assertIs(result, dmatrix)
        np.testing.assert_array_equal(
            result.get_label(), np.zeros(module.batch_size * module.output_dim)
        )
        self.assertEqual(module.training_n, dmatrix.num_row())

    def test_input_is_ndarray_training_true_dtrain_none(self):
        """Test with np.ndarray input, training mode True, and dtrain is None."""
        module = xgm.XGBModule(10, 5, 3)
        data = np.random.rand(10, 5)
        result = module._input_checking_setting(data)
        self.assertIs(result, module.dtrain)
        self.assertIsInstance(result, xgb.DMatrix)
        np.testing.assert_array_equal(
            result.get_label(), np.zeros(module.batch_size * module.output_dim)
        )
        self.assertEqual(module.training_n, data.shape[0])

    def test_input_is_dmatrix_training_true_dtrain_set_same_nrows(self):
        """Test with xgb.DMatrix input, training mode True, and dtrain already set with same number of rows."""
        module = xgm.XGBModule(10, 5, 3)
        data = np.random.rand(10, 5)
        dmatrix = xgb.DMatrix(data)
        result = module._input_checking_setting(dmatrix)
        self.assertIs(result, module.dtrain)
        self.assertIs(result, dmatrix)

    def test_input_is_dmatrix_training_true_dtrain_set_different_nrows(self):
        """Test with xgb.DMatrix input, training mode True, and dtrain already set with different number of rows."""
        module = xgm.XGBModule(10, 5, 3)
        data1 = np.random.rand(10, 5)
        module(data1)
        data2 = np.random.rand(5, 5)
        with self.assertRaises(AssertionError) as context:
            module(data2)
        self.assertIn(
            "Changing datasets while training is not currently supported",
            str(context.exception),
        )

    def test_input_is_dmatrix_training_false(self):
        """Test with xgb.DMatrix input and training mode False."""
        module = xgm.XGBModule(10, 5, 3)
        module(np.random.rand(10, 5))
        module.eval()
        data = np.random.rand(10, 5)
        dmatrix = xgb.DMatrix(data)
        result = module._input_checking_setting(dmatrix)
        self.assertIs(result, dmatrix)

    def test_input_is_ndarray_training_false(self):
        """Test with np.ndarray input and training mode False."""
        module = xgm.XGBModule(10, 5, 3)
        module(np.random.rand(10, 5))
        module.eval()
        data = np.random.rand(10, 5)
        result = module._input_checking_setting(data)
        self.assertIsInstance(result, xgb.DMatrix)
        self.assertEqual(result.num_row(), data.shape[0])

    def test_input_invalid_type(self):
        """Test with invalid input type."""
        module = xgm.XGBModule(10, 5, 3)
        data = [1, 2, 3]  # Invalid type
        with self.assertRaises(AssertionError):
            module._input_checking_setting(data)



from typing import Union

import numpy as np
import pandas as pd

from scipy.linalg import cho_solve, cho_factor
import torch
import torch.nn as nn

from gbnet.base import BaseGBModule


class GBLinear(BaseGBModule):
    """A linear gradient boosting module that uses gradient boosting for updates.

    This module implements a linear layer that can be trained using gradient boosting.
    It maintains state between iterations and updates parameters using computed gradients
    and hessians.

    Parameters
    ----------
    input_dim : int
        Input feature dimension
    output_dim : int
        Output prediction dimension
    bias : bool, optional
        Whether to include a bias term. Defaults to True.
    lr : float, optional
        Learning rate for parameter updates. Defaults to 0.5.
    min_hess : float, optional
        Minimum hessian threshold. Defaults to 0.0.
    lambd : float, optional
        L2 regularization parameter. Defaults to 0.01.

    Attributes
    ----------
    input_dim : int
        Input feature dimension
    output_dim : int
        Output prediction dimension
    min_hess : float
        Minimum hessian threshold
    bias : bool
        Whether bias term is included
    lr : float
        Learning rate
    lambd : float
        L2 regularization parameter
    linear : nn.Linear
        The underlying linear layer
    FX : torch.Tensor
        Current predictions tensor
    input : numpy.ndarray
        Input data cache
    g : torch.Tensor
        Gradient cache
    h : torch.Tensor
        Hessian cache
    """

    def __init__(
        self,
        input_dim,
        output_dim,
        bias=True,
        lr=0.5,
        min_hess=0.0,
        lambd=0.01,
    ):
        super(GBLinear, self).__init__()
        self.input_dim = input_dim
        self.output_dim = output_dim
        self.min_hess = min_hess
        self.bias = bias
        self.lr = lr
        self.lambd = lambd

        self.linear = nn.Linear(self.input_dim, self.output_dim, bias=self.bias)
        self.FX = None
        self.input = None
        self.g = None
        self.h = None

    def _input_checking_setting(self, x: Union[torch.Tensor, np.ndarray, pd.DataFrame]):
        assert isinstance(x, (torch.Tensor, np.ndarray, pd.DataFrame))

        if isinstance(x, np.ndarray):
            x = torch.Tensor(x)
        if isinstance(x, pd.DataFrame):
            x = torch.Tensor(np.array(x))

        if self.training:
            self.input = x.detach().numpy()  # TODO add input checks

        return x

    def forward(self, x: Union[torch.Tensor, np.ndarray, pd.DataFrame]):
        x = self._input_checking_setting(x)

        self.FX = self.linear(x)
        if self.training:
            self.FX.retain_grad()
        return self.FX

    def gb_calc(self):
        """Calculate gradients and stores in the object"""
        if self.FX is None or self.FX.grad is None:
            raise RuntimeError("Backward must be called before gb_step.")

        self.g, self.h = self._get_grad_hess_FX()

    def gb_step(self):
        """Uses stored gradients to update weights"""
        if self.g is None and self.h is None:
            self.gb_calc()

        with torch.no_grad():
            if self.bias:
                X = np.concatenate(
                    [np.ones([self.input.shape[0], 1]), self.input], axis=1
                )
            else:
                X = self.input

            updated_B = ridge_regression(
                X, (self.g / self.h).detach().numpy(), self.lambd
            )

            updated_weight_dir = updated_B[1:, :].T
            self.linear.weight -= self.lr * torch.Tensor(updated_weight_dir)

            if self.bias:
                updated_bias_dir = updated_B[0:1, :].flatten()
                self.linear.bias -= self.lr * torch.Tensor(updated_bias_dir)
        self.g = None
        self.h = None


def ridge_regression(X, y, lambd):
    """Solves ridge regression using Cholesky decomposition.

    Fastest method tested.

    Args:
        X (np.ndarray): Design matrix of shape (n_samples, n_features)
        y (np.ndarray): Target values of shape (n_samples,)
        lambd (float): Ridge regularization parameter

    Returns:
        np.ndarray: Fitted coefficients of shape (n_features,)

    The function solves the ridge regression problem:
    min_beta ||X beta - y||^2 + lambd ||beta||^2
    using the normal equations and Cholesky decomposition for numerical stability.
    """
    n, d = X.shape
    A = X.T @ X + lambd * np.eye(d)
    c = X.T @ y
    L = cho_factor(A)
    beta = cho_solve(L, c)
    return beta



from unittest import mock, TestCase

import lightgbm as lgb
import numpy as np
import pandas as pd
import torch
from gbnet import lgbmodule as lgm


def test_basic_loss():
    gbm = lgm.LGBModule(5, 3, 1, params={"min_data_in_leaf": 0})
    floss = torch.nn.MSELoss()

    gbm.zero_grad()
    np.random.seed(11010)
    input_dataset = lgb.Dataset(np.random.random([5, 3]))
    preds = gbm(input_dataset)
    loss = floss(preds.flatten(), torch.Tensor(np.array([1, 2, 3, 4, 5])).flatten())

    loss.backward(create_graph=True)

    m_obj = mock.MagicMock(side_effect=lgm.LightGBObj)
    m_train = mock.MagicMock(side_effect=lgb.train)
    with (
        mock.patch("gbnet.lgbmodule.LightGBObj", m_obj),
        mock.patch("lightgbm.train", m_train),
    ):
        gbm.gb_step()

    assert (
        np.max(
            np.abs(
                m_obj.call_args_list[-1].args[0].detach().numpy()
                - np.array([-2, -4, -6, -8, -10]).reshape([-1, 1])
            )
        )
        < 1e-8
    )

    assert (
        np.max(
            np.abs(
                m_obj.call_args_list[-1].args[1].detach().numpy()
                - np.array([2, 2, 2, 2, 2]).reshape([-1, 1])
            )
        )
        < 1e-8
    )

    m_train.assert_called_once()


def test_LightGBObj():
    np.random.seed(10101)
    grad = torch.tensor(np.random.random([20, 10]))
    hess = torch.tensor(np.random.random([20, 10]))

    obj = lgm.LightGBObj(grad, hess)

    ograd, ohess = obj(1, 2)
    assert (
        np.max(np.abs(ograd - grad.detach().numpy())) == 0
    ), "LightGBObj grad does not match instantiation"
    assert (
        np.max(np.abs(ohess - hess.detach().numpy())) == 0
    ), "LightGBObj hess does not match instantiation"


class TestLGBModule(TestCase):
    def test_input_is_dataset_training_true_train_dat_none(self):
        """Test with lgb.Dataset input, training mode True, and train_dat is None."""
        module = lgm.LGBModule(100, 10, 1)
        data = np.random.rand(100, 10)
        dataset = lgb.Dataset(data)
        result = module._input_checking_setting(dataset)
        self.assertIs(result, module.train_dat)
        self.assertIsInstance(result, lgb.Dataset)

    def test_input_is_ndarray_training_true_train_dat_none(self):
        """Test with np.ndarray input, training mode True, and train_dat is None."""
        module = lgm.LGBModule(100, 10, 1)
        data = np.random.rand(100, 10)
        result = module._input_checking_setting(data)
        self.assertIs(result, module.train_dat)
        self.assertIsInstance(result, lgb.Dataset)

    def test_input_is_dataframe_training_true_train_dat_none(self):
        """Test with pd.DataFrame input, training mode True, and train_dat is None."""
        module = lgm.LGBModule(100, 10, 1)
        data = pd.DataFrame(np.random.rand(100, 10))
        result = module._input_checking_setting(data)
        self.assertIs(result, module.train_dat)
        self.assertIsInstance(result, lgb.Dataset)

    def test_input_is_dataset_training_true_train_dat_set_same_nrows(self):
        """Test with lgb.Dataset input, training mode True, train_dat set, same number of data."""
        module = lgm.LGBModule(100, 10, 1)
        data = np.random.rand(100, 10)
        dataset = lgb.Dataset(data)
        result = module._input_checking_setting(dataset)
        self.assertIs(result, module.train_dat)
        self.assertIs(result, dataset)

    def test_input_is_dataset_training_false(self):
        """Test with lgb.Dataset input and training mode False."""
        module = lgm.LGBModule(100, 10, 1)
        module(np.random.rand(10, 10))
        module.eval()
        data = np.random.rand(100, 10)
        dataset = lgb.Dataset(data)
        result = module._input_checking_setting(dataset)
        # Should return the original data
        self.assertTrue(np.array_equal(result, data))
        # Ensure free_raw_data is set to False
        self.assertFalse(dataset.free_raw_data)

    def test_input_is_ndarray_training_false(self):
        """Test with np.ndarray input and training mode False."""
        module = lgm.LGBModule(100, 10, 1)
        module.eval()
        data = np.random.rand(100, 10)
        result = module._input_checking_setting(data)
        # Should return input data as is
        self.assertIs(result, data)

    def test_input_is_dataframe_training_false(self):
        """Test with pd.DataFrame input and training mode False."""
        module = lgm.LGBModule(100, 10, 1)
        module.eval()
        data = pd.DataFrame(np.random.rand(100, 10))
        result = module._input_checking_setting(data)
        # Should return input data as is
        self.assertIs(result, data)

    def test_input_invalid_type(self):
        """Test with invalid input type."""
        module = lgm.LGBModule(10, 5, 1)
        data = [1, 2, 3]  # Invalid type
        with self.assertRaises(AssertionError):
            module._input_checking_setting(data)

    def test_if_bst_is_none(self):
        """
        Test scenario where self.training = True, self.train_dat is not None,
        and self.bst is None. The method should return self.train_dat directly.
        """
        module = lgm.LGBModule(10, 5, 1)
        arr = np.random.rand(10, 5)
        module._set_train_dat(lgb.Dataset(arr))

        # Since bst is None, method should return train_dat directly
        result = module._input_checking_setting(np.random.rand(10, 5))
        self.assertIs(
            result, module.train_dat, "Expected to return train_dat when bst is None."
        )

    def test_raises_input_changed_lgb_dataset(self):
        """
        Test scenario where self.training = True, self.train_dat not None,
        self.bst not None, and input_dataset is an lgb.Dataset with a _handle.
        Expect no assertion error and return self.train_dat.
        """
        module = lgm.LGBModule(10, 5, 1)
        arr = np.random.rand(10, 5)
        initial_dataset = lgb.Dataset(arr)
        initial_dataset.construct()
        module._set_train_dat(initial_dataset)

        # bst is non-None now
        module.bst = object()

        with self.assertRaises(AssertionError):
            module._input_checking_setting(lgb.Dataset(np.random.random([10, 5])))

    def test_raises_input_changed_ndarray(self):
        """
        Test scenario where self.training = True, self.train_dat not None,
        self.bst not None, and input_dataset is an lgb.Dataset with a _handle.
        Expect no assertion error and return self.train_dat.
        """
        module = lgm.LGBModule(10, 5, 1)
        arr = np.random.rand(10, 5)
        initial_dataset = lgb.Dataset(arr)
        initial_dataset.construct()
        module._set_train_dat(initial_dataset)

        # bst is non-None now
        module.bst = object()

        with self.assertRaises(AssertionError):
            module._input_checking_setting(np.random.random([11, 5]))



from typing import Union
import lightgbm as lgb
import numpy as np
import pandas as pd
import torch
from torch import nn

from gbnet.base import BaseGBModule


class LGBModule(BaseGBModule):
    """LightGBM Module that wraps LightGBM boosting into a PyTorch Module.

    This module allows integration of LightGBM gradient boosting with PyTorch neural networks.
    It maintains the boosting model state and handles both training and inference.

    Args:
        batch_size (int): Size of training data
        input_dim (int): Dimension of input features
        output_dim (int): Dimension of output predictions
        params (dict, optional): Parameters passed to LightGBM. Defaults to {}.
        min_hess (float, optional): Minimum hessian value submitted to LightGBM. Defaults to 0.

    Attributes:
        batch_size (int): Size of mini-batches
        input_dim (int): Input feature dimension
        output_dim (int): Output prediction dimension
        params (dict): LightGBM parameters
        bst (lightgbm.Booster): The underlying LightGBM booster
        FX (torch.nn.Parameter): Current predictions tensor
        train_dat (lightgbm.Dataset): Training dataset used for caching
        min_hess (float): Minimum hessian threshold
    """

    def __init__(self, batch_size, input_dim, output_dim, params={}, min_hess=0):
        super(LGBModule, self).__init__()
        self.batch_size = batch_size
        self.input_dim = input_dim
        self.output_dim = output_dim

        self.params = params
        self.bst = None

        self.FX = nn.Parameter(
            torch.tensor(
                np.zeros([batch_size, output_dim]),
                dtype=torch.float,
            )
        )
        self.train_dat = None
        self.min_hess = min_hess
        self.grad = None
        self.hess = None

    def _set_train_dat(self, input_dataset: lgb.Dataset):
        if input_dataset.params is None:
            input_dataset.params = {"verbose": -1}
        else:
            input_dataset.params.update({"verbose": -1})
        input_dataset.free_raw_data = False
        self.train_dat = input_dataset

    def _input_checking_setting(
        self, input_dataset: Union[lgb.Dataset, np.ndarray, pd.DataFrame]
    ):
        assert isinstance(input_dataset, (lgb.Dataset, np.ndarray, pd.DataFrame))
        if self.training:
            if self.train_dat is None:
                self._set_train_dat(
                    input_dataset
                    if isinstance(input_dataset, lgb.Dataset)
                    else lgb.Dataset(input_dataset)
                )
            if self.bst is None:
                return self.train_dat
            if isinstance(input_dataset, lgb.Dataset):
                assert (
                    input_dataset._handle is not None
                ), "Changing datasets during training is not supported. If trying to do prediction, call LGBModule.eval() first."
            else:  # NEW
                assert (
                    self.batch_size == input_dataset.shape[0]
                ), "Changing datasets during training is not supported. If trying to do prediction, call LGBModule.eval() first."
            return self.train_dat

        if isinstance(input_dataset, lgb.Dataset):
            # Clunky way to get original data
            input_dataset.free_raw_data = False
            input_dataset.construct()
            return input_dataset.get_data()
        return input_dataset

    def forward(
        self,
        input_dataset: Union[lgb.Dataset, np.ndarray, pd.DataFrame],
        return_tensor=True,
    ):
        """Forward pass through the LightGBM module.

        Args:
            input_dataset (Union[lgb.Dataset, np.ndarray, pd.DataFrame]): Input data for prediction.
                Can be a LightGBM Dataset, numpy array, or pandas DataFrame.
            return_tensor (bool, optional): Whether to return predictions as a PyTorch tensor.
                Defaults to True.

        Returns:
            Union[torch.Tensor, np.ndarray]: Model predictions. Returns a PyTorch tensor if
            return_tensor=True, otherwise returns a numpy array.

        The forward pass handles both train and eval
        - In train mode, maintains state between iterations and updates internal FX tensor
        - In eval mode, generates predictions on new data using the trained model
        """
        input_dataset = self._input_checking_setting(input_dataset)

        # TODO figure out how actual batch training works here
        if self.training:
            if self.bst:
                preds = self.bst._Booster__inner_predict(0).copy()
            else:
                preds = np.zeros([self.batch_size, self.output_dim])
        else:
            if self.bst:
                preds = self.bst.predict(input_dataset).copy()
            else:
                preds = np.zeros(
                    [input_dataset.shape[0], self.output_dim], dtype=torch.float
                )

        if self.training:
            FX_detach = self.FX.detach()
            FX_detach.copy_(
                torch.tensor(
                    preds.reshape([self.batch_size, self.output_dim]), dtype=torch.float
                )
            )

        if return_tensor:
            if self.training:
                return self.FX
            else:
                return torch.tensor(
                    preds.reshape([-1, self.output_dim]), dtype=torch.float
                )
        return preds

    def gb_calc(self):
        self.grad, self.hess = self._get_grad_hess_FX()

    def gb_step(self):
        """Performs a gradient boosting step to update the model.

        This method:
        1. Computes gradients and hessians from the current predictions
        2. Creates a LightGBM objective using the computed gradients/hessians
        3. Updates the internal boosting model by either:
           - Updating the existing model if one exists
           - Training a new model for 1 boosting round if no model exists

        The gradients are scaled by batch size and hessians are clipped to a minimum value
        to ensure numerical stability.

        Returns:
            None
        """
        if self.grad is None and self.hess is None:
            self.gb_calc()

        obj = LightGBObj(self.grad, self.hess)
        input_params = self.params.copy()
        input_params.update(
            {
                "objective": obj,
                "num_class": self.output_dim,
                "verbose": -1,
                "verbosity": -1,
            }
        )

        if self.bst is not None:
            self.bst.update(train_set=self.train_dat, fobj=obj)
        else:
            self.bst = lgb.train(
                params=input_params,
                train_set=self.train_dat,
                num_boost_round=1,
                keep_training_booster=True,
            )
        self.grad = None
        self.hess = None


class LightGBObj:
    """Helper class for use with LightGBM as a backend for LGBModule"""

    def __init__(self, grad, hess):
        self.grad = grad.detach().numpy()
        self.hess = hess.detach().numpy()

    def __call__(self, y_true, y_pred):
        if self.grad.shape[1] > 1:
            return self.grad, self.hess
        return self.grad.flatten(), self.hess.flatten()



from typing import Union
import warnings
import numpy as np
import pandas as pd
import torch
import xgboost as xgb
from torch import nn

from gbnet.base import BaseGBModule


class XGBModule(BaseGBModule):
    """XGBoost Module that wraps XGBoost boosting into a PyTorch Module.

    This module allows integration of XGBoost gradient boosting with PyTorch neural networks.
    It maintains the boosting model state and handles both training and inference.

    Args:
        batch_size (int): Size of training data
        input_dim (int): Dimension of input features
        output_dim (int): Dimension of output predictions
        params (dict, optional): Parameters passed to LightGBM. Defaults to {}.
        min_hess (float, optional): Minimum hessian value submitted to LightGBM. Defaults to 0.

    Attributes:
        batch_size (int): Size of mini-batches
        input_dim (int): Input feature dimension
        output_dim (int): Output prediction dimension
        params (dict): LightGBM parameters
        bst (lightgbm.Booster): The underlying LightGBM booster
        FX (torch.nn.Parameter): Current predictions tensor
        train_dat (lightgbm.Dataset): Training dataset used for caching
        min_hess (float): Minimum hessian threshold
    """

    def __init__(self, batch_size, input_dim, output_dim, params={}, min_hess=0):
        super(XGBModule, self).__init__()
        self.batch_size = batch_size
        self.input_dim = input_dim
        self.output_dim = output_dim

        self.params = params
        self.params["objective"] = "reg:squarederror"
        self.params["base_score"] = 0
        self.n_completed_boost_rounds = 0
        self.min_hess = min_hess

        init_matrix = np.zeros([batch_size, input_dim])
        self.bst = xgb.train(
            self.params,
            xgb.DMatrix(init_matrix, label=np.zeros(batch_size * output_dim)),
            num_boost_round=0,
        )
        self.n_completed_boost_rounds = 0
        self.dtrain = None
        self.training_n = None

        self.FX = nn.Parameter(
            torch.tensor(
                np.zeros([batch_size, output_dim]),
                dtype=torch.float,
            )
        )

    def _check_training_data(self):
        if self.dtrain.get_weight().shape[0] > 0:
            warnings.warn(
                "Weights will not work properly when defined as part of the input DMatrix. Weights should be defined in the loss."
            )

    def _input_checking_setting(
        self, input_data: Union[xgb.DMatrix, pd.DataFrame, np.ndarray]
    ):
        assert isinstance(input_data, (xgb.DMatrix, pd.DataFrame, np.ndarray))
        if self.training:
            if self.dtrain is None:
                if isinstance(input_data, xgb.DMatrix):
                    input_data.set_label(np.zeros(self.batch_size * self.output_dim))
                    self.dtrain = input_data
                    self.training_n = input_data.num_row()
                    self._check_training_data()
                else:
                    self.dtrain = xgb.DMatrix(
                        input_data, label=np.zeros(self.batch_size * self.output_dim)
                    )
                    self.training_n = input_data.shape[0]
            compare_n = (
                input_data.num_row()
                if isinstance(input_data, xgb.DMatrix)
                else input_data.shape[0]
            )
            assert (
                compare_n == self.training_n
            ), "Changing datasets while training is not currently supported. If trying to make predictions, set Module to eval mode via `Module.eval()`"
            return self.dtrain
        return (
            input_data
            if isinstance(input_data, xgb.DMatrix)
            else xgb.DMatrix(input_data)
        )

    def forward(
        self,
        input_data: Union[xgb.DMatrix, np.ndarray, pd.DataFrame],
        return_tensor: bool = True,
    ):
        """Forward pass through the XGBoost module.

        Args:
            input_dataset (Union[xgb.DMatrix, np.ndarray, pd.DataFrame]): Input data for prediction.
                Can be a XGBoost DMatrix, numpy array, or pandas DataFrame.
            return_tensor (bool, optional): Whether to return predictions as a PyTorch tensor.
                Defaults to True.

        Returns:
            Union[torch.Tensor, np.ndarray]: Model predictions. Returns a PyTorch tensor if
            return_tensor=True, otherwise returns a numpy array.

        The forward pass handles both train and eval
        - In train mode, maintains state between iterations and updates internal FX tensor
        - In eval mode, generates predictions on new data using the trained model
        """
        input_data = self._input_checking_setting(input_data)
        preds = self.bst.predict(input_data)

        if self.training:
            FX_detach = self.FX.detach()
            FX_detach.copy_(
                torch.tensor(
                    preds.reshape([self.batch_size, self.output_dim]), dtype=torch.float
                )
            )

        if return_tensor:
            if self.training:
                return self.FX
            else:
                return torch.tensor(
                    preds.reshape([-1, self.output_dim]), dtype=torch.float
                )
        return preds

    def gb_calc(self):
        self.grad, self.hess = self._get_grad_hess_FX()

    def gb_step(self):
        """Performs a gradient boosting step to update the model.

        This method:
        1. Computes gradients and hessians from the current predictions
        3. Updates the internal boosting model

        The gradients are scaled by batch size and hessians are clipped to a minimum value
        to ensure numerical stability.

        Returns:
            None
        """
        if self.grad is None and self.hess is None:
            self.gb_calc()

        self._gb_step_grad_hess(self.grad, self.hess)
        self.grad = None
        self.hess = None

    def _gb_step_grad_hess(self, grad, hess):
        obj = XGBObj(grad, hess)
        g, h = obj(np.zeros([self.batch_size, self.output_dim]), None)

        if xgb.__version__ <= "2.0.3":
            self.bst.boost(
                self.dtrain,
                g,
                h,
            )
        else:
            self.bst.boost(
                self.dtrain,
                self.n_completed_boost_rounds + 1,
                g,
                h,
            )
        self.n_completed_boost_rounds = self.n_completed_boost_rounds + 1


class XGBObj:
    """Helper class for use with XGBoost as a backend for XGBModule"""

    def __init__(self, grad, hess):
        self.grad = grad
        self.hess = hess

    def __call__(self, preds, dtrain):
        if len(preds.shape) == 2:
            M = preds.shape[0]
            N = preds.shape[1]
        else:
            M = preds.shape[0]
            N = 1

        if xgb.__version__ >= "2.1.0":
            g = self.grad.detach().numpy().reshape([M, N])
            h = self.hess.detach().numpy().reshape([M, N])
        else:
            g = self.grad.detach().numpy().reshape([M * N, 1])
            h = self.hess.detach().numpy().reshape([M * N, 1])

        return g, h



import numpy as np
import pytest
import torch
import torch.nn as nn

from gbnet.gblinear import GBLinear, ridge_regression


def test_gblinear_init():
    model = GBLinear(input_dim=5, output_dim=2)
    assert model.input_dim == 5
    assert model.output_dim == 2
    assert model.bias
    assert model.lr == 0.5
    assert model.min_hess == 0.0
    assert model.lambd == 0.01
    assert isinstance(model.linear, nn.Linear)


def test_gblinear_forward():
    model = GBLinear(input_dim=3, output_dim=2)
    x = torch.randn(4, 3)

    out = model(x)
    assert out.shape == (4, 2)
    assert model.FX is not None
    assert model.input is not None


def test_gblinear_training_step():
    model = GBLinear(input_dim=3, output_dim=2)
    x = torch.randn(4, 3)
    y = torch.randn(4, 2)

    criterion = nn.MSELoss()

    # Forward pass
    out = model(x)
    loss = criterion(out, y)

    # Backward pass
    loss.backward(create_graph=True)

    # Store initial weights
    init_weights = model.linear.weight.clone()
    init_bias = model.linear.bias.clone()

    # GB step
    model.gb_calc()
    model.gb_step()

    # Check weights were updated
    assert not torch.allclose(model.linear.weight, init_weights)
    assert not torch.allclose(model.linear.bias, init_bias)


def test_ridge_regression():
    # Create synthetic data
    np.random.seed(42)
    X = np.random.randn(100, 5)
    true_beta = np.array([1, -0.5, 0.2, -0.3, 0.8])
    y = X @ true_beta + np.random.randn(100) * 0.1

    # Solve ridge regression
    lambd = 0.1
    beta = ridge_regression(X, y, lambd)

    # Check shape
    assert beta.shape == true_beta.shape

    # Check solution is reasonable (close to true parameters)
    assert np.allclose(beta, true_beta, atol=0.5)


def test_input_validation():
    model = GBLinear(input_dim=3, output_dim=2)

    # Test numpy input
    x_np = np.random.randn(4, 3)
    out = model(x_np)
    assert isinstance(out, torch.Tensor)
    assert out.shape == (4, 2)

    # Test pandas input
    import pandas as pd

    x_pd = pd.DataFrame(x_np)
    out = model(x_pd)
    assert isinstance(out, torch.Tensor)
    assert out.shape == (4, 2)

    # Test invalid input
    with pytest.raises(AssertionError):
        model("invalid input")


def test_gblinear_no_bias():
    model = GBLinear(input_dim=3, output_dim=2, bias=False)
    assert not model.bias
    assert model.linear.bias is None

    x = torch.randn(4, 3)
    out = model(x)
    assert out.shape == (4, 2)



import time

import lightgbm as lgb
import numpy as np
import torch
import xgboost as xgb

from gbnet import lgbmodule, xgbmodule


def test_readme_examples():
    # Generate Dataset
    np.random.seed(100)
    n = 1000
    input_dim = 20
    output_dim = 1
    X = np.random.random([n, input_dim])
    B = np.random.random([input_dim, output_dim])
    Y = X.dot(B) + np.random.random([n, output_dim])

    iters = 100
    t0 = time.time()

    # XGBoost training for comparison
    xbst = xgb.train(
        params={"objective": "reg:squarederror", "base_score": 0.0},
        dtrain=xgb.DMatrix(X, label=Y),
        num_boost_round=iters,
    )
    t1 = time.time()

    # LightGBM training for comparison
    lbst = lgb.train(
        params={"verbose": -1},
        train_set=lgb.Dataset(X, label=Y.flatten(), init_score=[0 for i in range(n)]),
        num_boost_round=iters,
    )
    t2 = time.time()

    # XGBModule training
    xnet = xgbmodule.XGBModule(n, input_dim, output_dim, params={})
    xmse = torch.nn.MSELoss()

    X_dmatrix = xgb.DMatrix(X)
    for i in range(iters):
        xnet.zero_grad()
        xpred = xnet(X_dmatrix)

        loss = 1 / 2 * xmse(xpred, torch.Tensor(Y))  # xgboost uses 1/2 (Y - P)^2
        loss.backward(create_graph=True)

        xnet.gb_step()
    xnet.eval()
    t3 = time.time()

    # LGBModule training
    lnet = lgbmodule.LGBModule(n, input_dim, output_dim, params={})
    lmse = torch.nn.MSELoss()

    X_dataset = lgb.Dataset(X)
    for i in range(iters):
        lnet.zero_grad()
        lpred = lnet(X_dataset)

        loss = lmse(lpred, torch.Tensor(Y))
        loss.backward(create_graph=True)

        lnet.gb_step()
    lnet.eval()
    t4 = time.time()

    assert np.isclose(
        0.0,
        np.max(
            np.abs(
                xbst.predict(xgb.DMatrix(X))
                - xnet(X_dmatrix).detach().numpy().flatten()
            )
        ),
        atol=1e-05,
    )
    assert np.isclose(
        0.0,
        np.max(np.abs(lbst.predict(X) - lnet(X).detach().numpy().flatten())),
        atol=1e-05,
    )
    assert (t3 - t2) / (t1 - t0) < 4
    assert (t4 - t3) / (t2 - t1) < 4


def test_combine_example():
    # Create new module that jointly trains multi-output xgboost and lightgbm models
    # the outputs of these gbm models is then combined by a linear layer
    class GBPlus(torch.nn.Module):
        def __init__(self, input_dim, intermediate_dim, output_dim):
            super(GBPlus, self).__init__()

            self.xgb = xgbmodule.XGBModule(n, input_dim, intermediate_dim, {"eta": 0.1})
            self.lgb = lgbmodule.LGBModule(n, input_dim, intermediate_dim, {"eta": 0.1})
            self.linear = torch.nn.Linear(intermediate_dim, output_dim)

        def forward(self, input_array):
            xpreds = self.xgb(input_array)
            lpreds = self.lgb(input_array)
            preds = self.linear(xpreds + lpreds)
            return preds

        def gb_step(self):
            self.xgb.gb_step()
            self.lgb.gb_step()

    # Generate Dataset
    np.random.seed(100)
    n = 1000
    input_dim = 10
    output_dim = 1
    X = np.random.random([n, input_dim])
    B = np.random.random([input_dim, output_dim])
    Y = X.dot(B) + np.random.random([n, output_dim])

    intermediate_dim = 10
    gbp = GBPlus(input_dim, intermediate_dim, output_dim)
    mse = torch.nn.MSELoss()
    optimizer = torch.optim.Adam(gbp.parameters(), lr=0.005)

    t0 = time.time()
    losses = []
    for i in range(100):
        optimizer.zero_grad()
        preds = gbp(X)

        loss = mse(preds, torch.Tensor(Y))
        loss.backward(create_graph=True)  # create_graph=True required for any gbnet
        losses.append(loss.detach().numpy().copy())

        gbp.gb_step()  # required to update the gbms
        optimizer.step()
    t1 = time.time()
    assert losses[-1] < 1e-07
    assert (t1 - t0) < 20



from setuptools import find_packages, setup

setup(
    name="gbnet",
    version="0.5.0",
    author="Michael Horrell",
    author_email="mthorrell@github.com",
    description="Gradient boosting libraries integrated with pytorch",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    url="https://github.com/mthorrell/gbnet",
    packages=find_packages(),
    install_requires=[
        "numpy",
        "torch",
    ],
)



import abc
import torch
from torch import nn


class BaseGBModule(nn.Module, abc.ABC):
    """Base class for gradient boosting modules.

    This abstract base class defines the common interface and functionality that all
    gradient boosting modules should implement.

    Attributes:
        min_hess (float) : minimum hessian value
    """

    def __init__(self, min_hess=0.0):
        super(BaseGBModule, self).__init__()
        self.min_hess = 0.0
        self.grad = None
        self.hess = None

    @abc.abstractmethod
    def _input_checking_setting(self, input_data):
        """Validate and prepare input data.

        Args:
            input_data: Input data in model-specific format

        Returns:
            Processed input data ready for model
        """
        pass

    @abc.abstractmethod
    def forward(self, input_data, return_tensor: bool = True):
        """Forward pass through the model.

        Args:
            input_data: Input data in model-specific format
            return_tensor: Whether to return predictions as PyTorch tensor

        Returns:
            Model predictions as tensor or numpy array
        """
        pass

    def _get_grad_hess_FX(self):
        grad = self.FX.grad * self.FX.shape[0]

        # parameters are independent row by row, so we can
        # at least calculate hessians column by column by
        # considering the sum of the gradient columns
        hesses = []
        for i in range(self.output_dim):
            hesses.append(
                torch.autograd.grad(grad[:, i].sum(), self.FX, retain_graph=True)[0][
                    :, i : (i + 1)
                ]
            )
        hess = torch.maximum(torch.cat(hesses, axis=1), torch.Tensor([self.min_hess]))
        return grad, hess

    @abc.abstractmethod
    def gb_step(self):
        """Perform one gradient boosting step.

        This method should implement the logic for:
        1. Getting gradients/hessians
        2. Training one boosting iteration
        3. Updating predictions
        """
        pass



import numpy as np
import torch
from unittest import TestCase
from gbnet.base import BaseGBModule


class TestBaseGBModule(TestCase):
    def test_get_grad_hess_FX(self):
        """Test gradient and hessian calculation"""

        class DummyGBModule(BaseGBModule):
            def _input_checking_setting(self, input_data):
                pass

            def forward(self, x):
                pass

            def gb_step(self):
                pass

        # Create dummy module
        module = DummyGBModule()
        module.batch_size = 3
        module.output_dim = 1
        module.min_hess = 1e-6

        # Set up test data
        module.FX = torch.nn.Parameter(torch.tensor([[1.0], [3.0], [5.0]]))
        target = torch.tensor([[0.5], [2.5], [4.5]])
        loss_fn = torch.nn.MSELoss()

        # Calculate loss and backward
        loss = loss_fn(module.FX, target)
        loss.backward(create_graph=True)

        # Get gradients and hessians
        grad, hess = module._get_grad_hess_FX()

        # Check shapes
        self.assertEqual(grad.shape, (3, 1))
        self.assertEqual(hess.shape, (3, 1))

        # Check gradient values (should be 2*(pred - target))
        expected_grad = 2 * (module.FX.data - target)
        np.testing.assert_array_almost_equal(
            grad.detach().numpy(), expected_grad.numpy()
        )

        # Check hessian values (should be 2 for MSE loss)
        expected_hess = torch.full_like(module.FX.data, 2.0)
        np.testing.assert_array_almost_equal(
            hess.detach().numpy(), expected_hess.numpy()
        )

    def test_min_hess_clipping(self):
        """Test minimum hessian clipping"""

        class DummyGBModule(BaseGBModule):
            def _input_checking_setting(self, input_data):
                pass

            def forward(self, x):
                pass

            def gb_step(self):
                pass

        module = DummyGBModule()
        module.batch_size = 2
        module.output_dim = 1
        module.min_hess = 1.5

        # Set up data that would produce hessians < min_hess
        module.FX = torch.nn.Parameter(torch.tensor([[1.0], [2.0]]))
        target = torch.tensor([[1.0], [2.0]])
        loss_fn = torch.nn.MSELoss()

        loss = loss_fn(module.FX, target)
        loss.backward(create_graph=True)

        _, hess = module._get_grad_hess_FX()

        # All hessians should be clipped to min_hess
        self.assertTrue(torch.all(hess >= module.min_hess))
        np.testing.assert_array_almost_equal(
            hess.detach().numpy(),
            np.full((2, 1), 2.0),  # MSE loss has constant hessian of 2.0
        )
